#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/xpm.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include "common.h"

#ifdef FRENCH
#include "french.h"
#else
#include "english.h"
#endif

#include "xquarto.h"
#include "sn.h"  /* Normal pixmaps */
#include "sf.h"  /* Faded pixmaps */
#include "sh.h"  /* Highlighted pixmaps */

static void CreateWidgets();
static void CreateWidgets2();
static void PrepareCallbacks();
static void PrepareCallbacks2();
static void ReadPixmaps();
static void ReadPixmaps2();
static void SetHints();
static void XFlushDpys();
static void XBlankPos(int i);
static void XBlankPiece(int i);
static void XInitPiece(int i);
static void XBlankOutPiece();
static void XFadePiece(int i);
static void XPutOutPiece(int i);
static void XAddPiece(int inpiece, int outpos);
static void XHilightPos(int i);
static void RandomPivots();
static void SetMessages(int i, int j);
static void ShowStatus();
static void InitGame();
static void PosProc();
static void PieceProc();
static void HilightProc();
static void GametypeProc();
static void LevelProc();
static void GeometryProc();
static void StartHumanProc();
static void StartComputerProc();
static void StartTwohumansProc();
static void StartNetworkProc();
static void CloseNetworkProc();
static void StartNetworkProc2();
static void MessageProc();
static void HelpProc();
static void CloseHelpProc();
static void HelpProc2();
static void CloseHelpProc2();
static void QuitProc();
static void QuitProc2();
static void NewProc2();
static void SaveProc();
static void LoadProc();
static void HintProc();
static void UndoProc();

Display       *dpy, *dpy2=NULL;
XtAppContext  app_cont;
XWMHints      *hints;
static char   host[256]="localhost:0.0";

Widget        top=NULL;
Widget        topbox;
Widget        menubox;
Widget        actions, actions_button;
Widget        start_button_human,start_button_computer,
              start_button_twohumans, start_button_network;
Widget        options, options_button, options_line1, options_line2;
Widget        undo_button, save_button, load_button, help_button;
Widget        helptop=NULL, helpdismiss;
Widget        networktop=NULL, networksend, networkdismiss, 
              networklabel, networkmsg, networkmain;
Widget        hint_button, quit_button;
Widget        gametype_button[NGAMETYPES];
Widget        level_button[NLEVELS];
Widget        geometry_button[NGEOMETRIES];
Widget        statustext, turntext, messagetext;
Widget        statusform, posform, pieceform, outputform, outpieceform;
Widget        pos[BOARDSIZE],piece[NPIECES],outpiece_command;

Widget        top2=NULL, topbox2;
Widget        actions2, actions_button2;
Widget        helptop2=NULL, helpdismiss2;
Widget        networktop2=NULL, networksend2, networkdismiss2, 
              networklabel2, networkmsg2, networkmain2;
Widget        new_button2, quit_button2;
Widget        help_button2, undo_button2, hint_button2;
Widget        statustext2, turntext2, messagetext2;
Widget        statusform2, posform2, pieceform2, outputform2, outpieceform2;
Widget        pos2[BOARDSIZE],piece2[NPIECES],outpiece_command2;

Pixel         white, white2;
Pixmap        piece_normal_pixmap[NPIECES],piece_normal_pixmap2[NPIECES];
Pixmap        piece_highlight_pixmap[NPIECES],piece_highlight_pixmap2[NPIECES];
Pixmap        piece_fade_pixmap[NPIECES],piece_fade_pixmap2[NPIECES];
PosPtr*       posp;
PiecePtr*     piecep;
GametypePtr*  gametypep;
LevelPtr*     levelp;
GeometryPtr*  geometryp;

PopupDefs     HelpDefs, AlertDefs;

int           realstart = 1;
int           proposedpiece;
int           maxdepth;
int           quarto_selected = 1;
int           move_num;
int           geometry;
int           firstplayer;
int           gamestatus;
int           gametype;

int	      seqpos[BOARDSIZE];
int           seqpiece[BOARDSIZE];
char          savefile[256];
char          statusline[80];

int	      board[FULLBOARDSIZE];
int           npiece[NPIECESIZE];

int           pivotpos[BOARDSIZE];
int           pivotpiece[NPIECES];

static char   message_body[256];

int main(int argc, char **argv)
{
  XOpenDisplay(NULL);

  sprintf(savefile,"%s/.xquarto_save", getenv("HOME"));

  gametype=QUARTO;
  geometry=NORMAL;
  firstplayer=HUMAN;
  maxdepth=3;

  HelpDefs.title = help_title;
  HelpDefs.string = help_string; 
  HelpDefs.distance = 457;
  HelpDefs.width = 523;
  HelpDefs.height = 278;
  AlertDefs.title = alert[0];
  AlertDefs.string = message_body;    
  AlertDefs.distance = 407;
  AlertDefs.width = 473;
  AlertDefs.height = 56;

  XtSetLanguageProc(NULL, NULL, NULL);

  top = XtVaOpenApplication(&app_cont, main_title, NULL, 0, &argc, argv,
			    fallback_resources, sessionShellWidgetClass, NULL);
  if (!top) exit(-1);

  CreateWidgets();
  PrepareCallbacks();
  
  XtRealizeWidget(top);
  ReadPixmaps();
  InitGame();
  SetHints();
  XtAppMainLoop(app_cont);
  return 0;
}

static void CreateWidgets()
{
  int i;

  topbox=XtVaCreateManagedWidget("topbox", boxWidgetClass, top, NULL);
  dpy = XtDisplay(topbox);

  actions_button=XtVaCreateManagedWidget("Run",menuButtonWidgetClass,
      topbox, NULL);
  actions=XtVaCreatePopupShell("menu",simpleMenuWidgetClass,
      actions_button, NULL);
  start_button_human=XtVaCreateManagedWidget("human",
      smeBSBObjectClass,actions,NULL);
  start_button_computer=XtVaCreateManagedWidget("computer",
      smeBSBObjectClass,actions,NULL);
  start_button_twohumans=XtVaCreateManagedWidget("twohumans",
      smeBSBObjectClass,actions,NULL);
  start_button_network=XtVaCreateManagedWidget("network",
      smeBSBObjectClass,actions,NULL);
  save_button=XtVaCreateManagedWidget("save",smeBSBObjectClass,actions,NULL);
  load_button=XtVaCreateManagedWidget("load",smeBSBObjectClass,actions,NULL);
  help_button=XtVaCreateManagedWidget("help",smeBSBObjectClass,actions,NULL);
  quit_button=XtVaCreateManagedWidget("quit",smeBSBObjectClass,actions,NULL);

  options_button=XtVaCreateManagedWidget("Options",menuButtonWidgetClass,
      topbox,NULL);
  options=XtVaCreatePopupShell("menu",simpleMenuWidgetClass,
      options_button,NULL);

  hint_button=XtVaCreateManagedWidget("Hint",commandWidgetClass,topbox,NULL);
  undo_button=XtVaCreateManagedWidget("Undo",commandWidgetClass,topbox,NULL);

  for (i=0; i<NGAMETYPES; i++)
    gametype_button[i]=XtVaCreateManagedWidget(gametype_name[i], 
         smeBSBObjectClass, options, NULL);
  options_line1=XtVaCreateManagedWidget("options_line1", 
         smeLineObjectClass, options, NULL);
  for (i=0; i<NLEVELS; i++)
    level_button[i]=XtVaCreateManagedWidget(level_name[i],
         smeBSBObjectClass, options, NULL);
  options_line2=XtVaCreateManagedWidget("options_line2",
         smeLineObjectClass, options, NULL);
  for (i=0; i<NGEOMETRIES; i++)
    geometry_button[i]=XtVaCreateManagedWidget(geometry_name[i],
         smeBSBObjectClass, options, NULL);

  statusform=XtVaCreateManagedWidget("statusform",formWidgetClass, topbox,
          XtNwidth, 178, XtNheight, 14,  XtNborderWidth, 0, NULL);
  statustext=XtVaCreateManagedWidget("status",labelWidgetClass, statusform,
          XtNwidth, 178, XtNheight, 14, 
          XtNhorizDistance, 2, XtNborderWidth, 0, NULL);

  posform=XtVaCreateManagedWidget("posform", formWidgetClass,topbox, NULL);
  for (i=0;i<BOARDSIZE;i++)
    pos[i] = XtVaCreateManagedWidget(pos_name[i], commandWidgetClass, 
                  posform, XtNlabel, "", NULL);

  outputform=XtVaCreateManagedWidget("output", formWidgetClass, topbox, NULL);
  turntext=XtVaCreateManagedWidget("turn", asciiTextWidgetClass,
        outputform, XtNdisplayCaret, False, NULL);
  messagetext=XtVaCreateManagedWidget("message", asciiTextWidgetClass,
        outputform, XtNdisplayCaret, False, NULL);

  outpieceform=XtVaCreateManagedWidget("outpiece",formWidgetClass, 
        outputform, NULL);
  outpiece_command=XtVaCreateManagedWidget("outpiece_cmd",commandWidgetClass, 
        outpieceform, XtNlabel, "", XtNsensitive, False, NULL);

  pieceform=XtVaCreateManagedWidget("pieceform",formWidgetClass,topbox,NULL);
  for (i=0;i<NPIECES;i++)
    piece[i] = XtVaCreateManagedWidget(piece_name[i], commandWidgetClass,
        pieceform, XtNlabel, "", NULL);
}

static void CreateWidgets2()
{
  static int nargc;
  static char **nargv;
  char *hostp;
  int i;

  nargc=0;
  XtVaGetValues(networkmsg, XtNstring, &hostp, NULL);
  i=strlen(hostp)-1;
  while (i>=0 && (isspace(hostp[i]) || iscntrl(hostp[i]))) {
    hostp[i]='\0';
    --i;
  }
  for(i=0; i<strlen(hostp); i++) if (hostp[i]=='\n') hostp[i]=' ';

  dpy2 = XtOpenDisplay(XtWidgetToApplicationContext(topbox),
     hostp, NULL, NULL, NULL, 0, &nargc, nargv);

  if (dpy2==NULL) {
    sprintf(message_body, alert[4], hostp);
    HelpProc(NULL, (XtPointer)&AlertDefs, NULL);
    return;
  }
  strncpy(host,hostp,255);

  XtVaSetValues(networklabel, XtNlabel, network_msg, NULL);
  XtVaSetValues(networksend, XtNlabel, network_send, NULL);
  XtVaSetValues(networkmsg, XtNstring, "", NULL);

  top2 = XtVaAppCreateShell(main_title, main_title2,
        applicationShellWidgetClass, dpy2, NULL);
  XtVaSetValues(top2, XtNtitle,  main_title2, NULL);
  topbox2=XtVaCreateManagedWidget("topbox", boxWidgetClass,top2,NULL);

  actions_button2=XtVaCreateManagedWidget("run",menuButtonWidgetClass,
      topbox2, NULL);
  actions2=XtVaCreatePopupShell("menu",simpleMenuWidgetClass,
      actions_button2, NULL);
  new_button2=XtVaCreateManagedWidget("new",smeBSBObjectClass,actions2,NULL);
  quit_button2=XtVaCreateManagedWidget("quit",smeBSBObjectClass,actions2,NULL);

  help_button2=XtVaCreateManagedWidget("help",commandWidgetClass,topbox2,NULL);
  hint_button2=XtVaCreateManagedWidget("hint",commandWidgetClass,topbox2,NULL);
  undo_button2=XtVaCreateManagedWidget("undo",commandWidgetClass,topbox2,NULL);

  statusform2=XtVaCreateManagedWidget("statusform",formWidgetClass, topbox2,
          XtNwidth, 178, XtNheight, 14, XtNborderWidth, 0, NULL);
  statustext2=XtVaCreateManagedWidget("status",labelWidgetClass, statusform2,
          XtNwidth, 178, XtNheight, 14, 
          XtNhorizDistance, 2, XtNborderWidth, 0, NULL);

  posform2=XtVaCreateManagedWidget("posform", formWidgetClass,topbox2, NULL);
  XtVaSetValues(posform2, XtNwidth, 200, XtNheight, 200, NULL);

  for (i=0;i<BOARDSIZE;i++)
    pos2[i] = XtVaCreateManagedWidget(pos_name[i], commandWidgetClass, 
    posform2, XtNlabel, "", NULL);

  outputform2=XtVaCreateManagedWidget("output", formWidgetClass, topbox2,NULL);
  turntext2=XtVaCreateManagedWidget("turn", asciiTextWidgetClass,
        outputform2, XtNdisplayCaret, False, NULL);
  messagetext2=XtVaCreateManagedWidget("message", asciiTextWidgetClass,
        outputform2, XtNdisplayCaret, False, NULL);

  outpieceform2=XtVaCreateManagedWidget("outpiece",formWidgetClass, 
        outputform2, NULL);
  outpiece_command2=XtVaCreateManagedWidget("outpiece_cmd",commandWidgetClass, outpieceform2, XtNlabel, "", XtNsensitive, False, NULL);

  pieceform2=XtVaCreateManagedWidget("pieceform",formWidgetClass,topbox2,NULL);
  for (i=0;i<NPIECES;i++)
    piece2[i] = XtVaCreateManagedWidget(piece_name[i],commandWidgetClass,pieceform2, XtNlabel, "", NULL);


  XtRealizeWidget(top2);
  XSetWMHints(dpy2, XtWindow(top2), hints);
  ReadPixmaps2();
}

void PrepareCallbacks()
{  
  int i;

  posp=(PosPtr*)calloc(BOARDSIZE,sizeof(PosPtr));
  piecep=(PiecePtr*)calloc(NPIECES,sizeof(PiecePtr)); 
  gametypep=(GametypePtr*)calloc(NGAMETYPES,sizeof(GametypePtr)); 
  levelp=(LevelPtr*)calloc(NLEVELS,sizeof(LevelPtr)); 
  geometryp=(GeometryPtr*)calloc(NGEOMETRIES,sizeof(GeometryPtr)); 
  
  for (i=0; i<BOARDSIZE; i++) {      /* Prepare for argument passing */
    *posp=(PosPtr)calloc(1,sizeof(PosData));
    (*posp)->posno=i;
    (*posp)->outpiece=&proposedpiece;
    XtAddCallback(pos[i], XtNcallback, PosProc, (XtPointer)(*posp));
    posp++;
  }
  
  for (i=0; i<NPIECES; i++) {       /* Prepare for argument passing */
    *piecep=(PiecePtr)calloc(1,sizeof(PieceData));
    (*piecep)->pieceno=i;
    (*piecep)->outpiece=&proposedpiece;
    XtAddCallback(piece[i], XtNcallback, PieceProc, (XtPointer)(*piecep));
    piecep++;
  }

  for (i=0; i<NGAMETYPES; i++) {       /* Prepare for argument passing */
    *gametypep=(GametypePtr)calloc(1,sizeof(GametypeData));
    (*gametypep)->gametype=i;
    XtAddCallback(gametype_button[i],XtNcallback,GametypeProc,(XtPointer)(*gametypep));
    gametypep++;
  }

  for (i=0; i<NLEVELS; i++) {       /* Prepare for argument passing */
    *levelp=(LevelPtr)calloc(1,sizeof(LevelData));
    (*levelp)->level=i;
    XtAddCallback(level_button[i],XtNcallback,LevelProc,(XtPointer)(*levelp));
    levelp++;
  }
  
  for (i=0; i<NGEOMETRIES; i++) {       /* Prepare for argument passing */
    *geometryp=(GeometryPtr)calloc(1,sizeof(GeometryData));
    (*geometryp)->geometry=i;
    XtAddCallback(geometry_button[i],XtNcallback,GeometryProc,(XtPointer)(*geometryp));
    geometryp++;
  }

  XtAddCallback(start_button_human, XtNcallback,StartHumanProc,(XtPointer)NULL);
  XtAddCallback(start_button_computer, XtNcallback,StartComputerProc,(XtPointer)NULL);
  XtAddCallback(start_button_twohumans, XtNcallback,StartTwohumansProc,(XtPointer)NULL);
  XtAddCallback(start_button_network, XtNcallback,StartNetworkProc,(XtPointer)NULL);
  XtAddCallback(save_button, XtNcallback, SaveProc, (XtPointer)NULL);
  XtAddCallback(load_button, XtNcallback, LoadProc, (XtPointer)NULL);
  XtAddCallback(hint_button, XtNcallback, HintProc, (XtPointer)NULL);
  XtAddCallback(help_button, XtNcallback, HelpProc, (XtPointer)&HelpDefs);
  XtAddCallback(quit_button, XtNcallback, QuitProc, (XtPointer)NULL);
  XtAddCallback(undo_button, XtNcallback, UndoProc,(XtPointer)NULL);
}

void PrepareCallbacks2()
{  
int i;

  XtAddCallback(new_button2, XtNcallback,  NewProc2,  (XtPointer)NULL);
  XtAddCallback(quit_button2, XtNcallback, QuitProc2, (XtPointer)NULL);
  XtAddCallback(help_button2, XtNcallback, HelpProc2, (XtPointer)&HelpDefs);
  XtAddCallback(hint_button2, XtNcallback, HintProc, (XtPointer)NULL);
  XtAddCallback(undo_button2, XtNcallback, UndoProc,(XtPointer)NULL);

  posp -= BOARDSIZE;
  for (i=0; i<BOARDSIZE; i++) {     /* posp pointers already defined */
    XtAddCallback(pos2[i], XtNcallback, PosProc, (XtPointer)(*posp));
    posp++;
  }

  piecep -= NPIECES;
  for (i=0; i<NPIECES; i++) {       /* piecep pointers already defined */
    XtAddCallback(piece2[i], XtNcallback, PieceProc, (XtPointer)(*piecep));
    piecep++;
  }
}

static void ReadPixmaps()
{
  XpmAttributes xpm_attributes;
  int val,i;

  xpm_attributes.valuemask = XpmSize|XpmReturnPixels|XpmColormap|XpmCloseness;
  xpm_attributes.colormap = XDefaultColormap(dpy,DefaultScreen(dpy));
  white = WhitePixel(dpy,DefaultScreen(dpy));
  xpm_attributes.closeness = 40000;
  for (i=0; i<NPIECES; i++) {
    val=XpmCreatePixmapFromData(XtDisplay(piece[i]),XtWindow(piece[i]),snp_xpm[i],&piece_normal_pixmap[i],NULL,&xpm_attributes);
    val=XpmCreatePixmapFromData(XtDisplay(piece[i]),XtWindow(piece[i]),shp_xpm[i],&piece_highlight_pixmap[i],NULL,&xpm_attributes);
    val=XpmCreatePixmapFromData(XtDisplay(piece[i]),XtWindow(piece[i]),sfp_xpm[i],&piece_fade_pixmap[i],NULL,&xpm_attributes);
  }
}

static void ReadPixmaps2()
{
  XpmAttributes xpm_attributes;
  int val,i;

  xpm_attributes.valuemask = XpmSize|XpmReturnPixels|XpmColormap|XpmCloseness;
  xpm_attributes.colormap = XDefaultColormap(dpy2,DefaultScreen(dpy2));
  white2 = WhitePixel(dpy2,DefaultScreen(dpy2));
  xpm_attributes.closeness = 40000;
  for (i=0; i<NPIECES; i++) {
    val=XpmCreatePixmapFromData(XtDisplay(piece2[i]),XtWindow(piece2[i]),snp_xpm[i],&piece_normal_pixmap2[i],NULL,&xpm_attributes);
    val=XpmCreatePixmapFromData(XtDisplay(piece2[i]),XtWindow(piece2[i]),shp_xpm[i],&piece_highlight_pixmap2[i],NULL,&xpm_attributes);
    val=XpmCreatePixmapFromData(XtDisplay(piece2[i]),XtWindow(piece2[i]),sfp_xpm[i],&piece_fade_pixmap2[i],NULL,&xpm_attributes);
  }
}

static void SetHints()
{
  hints = XGetWMHints(dpy, XtWindow(top));
  hints->flags=AllHints;
  hints->input=True;
  XSetWMHints(dpy, XtWindow(top), hints);
  XSetInputFocus(dpy,XtWindow(top),RevertToNone,CurrentTime);
}

static void XFlushDpys()
{
  XFlush(dpy);
  if (dpy2) XFlush(dpy2);
}
static void XBlankPos(int i)
{
  XtVaSetValues(pos[i], XtNlabel, "", XtNbackground, white,
     XtNbackgroundPixmap, XtUnspecifiedPixmap, NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(pos2[i], XtNlabel, "", XtNbackground, white2,
     XtNbackgroundPixmap, XtUnspecifiedPixmap, NULL);
}

static void XBlankPiece(int i)
{
  XtVaSetValues(piece[i], XtNbackground, white,
     XtNbackgroundPixmap, XtUnspecifiedPixmap, NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(piece2[i], XtNbackground, white2,
     XtNbackgroundPixmap, XtUnspecifiedPixmap, NULL);
}

static void XInitPiece(int i)
{
  XtVaSetValues(piece[i], XtNlabel, "", XtNsensitive, True,
     XtNbackgroundPixmap, piece_normal_pixmap[i], NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(piece2[i], XtNlabel, "", XtNsensitive, True,
     XtNbackgroundPixmap, piece_normal_pixmap2[i], NULL);
}

static void XBlankOutPiece()
{
  XtVaSetValues(outpiece_command, 
     XtNbackgroundPixmap, XtUnspecifiedPixmap, XtNbackground, white, NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(outpiece_command2, 
     XtNbackgroundPixmap, XtUnspecifiedPixmap, XtNbackground, white2, NULL);
}
 
static void XFadePiece(int i)
{
  XtVaSetValues(piece[i], XtNbackgroundPixmap, piece_fade_pixmap[i],
     XtNsensitive, False, NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(piece2[i], XtNbackgroundPixmap, piece_fade_pixmap2[i],
     XtNsensitive, False, NULL);
}

static void XPutOutPiece(int i)
{
  XtVaSetValues(outpiece_command,XtNbackgroundPixmap,
     piece_normal_pixmap[i],NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(outpiece_command2,XtNbackgroundPixmap,
     piece_normal_pixmap2[i],NULL);
  XFadePiece(i);
}

static void XAddPiece(int inpiece, int outpos)
{
  XtVaSetValues(pos[outpos], XtNbackgroundPixmap, 
     piece_normal_pixmap[inpiece],NULL); 
  if (firstplayer>=NETWORK1)
  XtVaSetValues(pos2[outpos], XtNbackgroundPixmap, 
     piece_normal_pixmap2[inpiece],NULL); 
}

static void XHilightPos(int i)
{
  XtVaSetValues(pos[i], XtNbackgroundPixmap,
     piece_highlight_pixmap[board[i]>>1], NULL);
  if (firstplayer>=NETWORK1)
  XtVaSetValues(pos2[i], XtNbackgroundPixmap,
     piece_highlight_pixmap2[board[i]>>1], NULL);
}

static void RandomPivots()
{
int i, j;

  srand(time(NULL));          /* generate random pivot position vector */
  for (i=0; i<BOARDSIZE; i++) /* (to make computer less predictable) */
    pivotpos[i]=-1;
  for (i=0; i<BOARDSIZE; ) {
    j=rand()%BOARDSIZE;
    if (pivotpos[j]==-1)
      pivotpos[j]=i++;
  }
                            /* generate random pivot piece vector */
  for (i=0; i<NPIECES; i++) /* (to make computer less predictable) */
    pivotpiece[i]=-1;
  for (i=0; i<NPIECES; ) {
    j=rand()%NPIECES;
    if (pivotpiece[j]==-1)
      pivotpiece[j]=i++;
  }
}

static void SetMessages(int i, int j)
{
  if (i>=0) XtVaSetValues(turntext, XtNstring, messages[i], NULL);
  if (j>=0) XtVaSetValues(messagetext, XtNstring, messages[j], NULL);
  if (firstplayer>=NETWORK1) {
    if (i>=0) XtVaSetValues(turntext2, XtNstring, messages[i], NULL);
    if (j>=0) XtVaSetValues(messagetext2, XtNstring, messages[j], NULL);
  }
}

static void ShowStatus()
{
  *statusline='\0';

  switch(firstplayer) {
    case HUMAN:     strcat(statusline,"H:C/"); break;
    case COMPUTER:  strcat(statusline,"C:H/"); break;
    case TWOHUMANS: strcat(statusline,"H:H/"); break;
    case NETWORK1:  strcat(statusline,"H:N/"); break;
    case NETWORK2:  strcat(statusline,"N:H/"); break;
    default:   break;
  }

  if (quarto_selected)
    strcat(statusline, "quarto/");
  else
    strcat(statusline, "tic/");

  strcat(statusline,"level /");
  statusline[strlen(statusline)-2]=maxdepth+'0';

  if (geometry==NORMAL)
    strcat(statusline,"normal");
  else if(geometry==NODIAGS)
    strcat(statusline,"nodiags");
  else
    strcat(statusline,"torus");

  XtVaSetValues(statustext, XtNlabel, statusline, NULL);

  if (firstplayer>=NETWORK1)
  XtVaSetValues(statustext2, XtNlabel, statusline, NULL);
}

static void InitGame()
{
  int i;

  move_num=0;

  for (i=0; i<BOARDSIZE; i++)   /* initialize board */
    board[i]=0;
  board[NPIECEPOS]=0;

  for (i=0; i<NPIECES; i++)   /* initialize pieces */
    npiece[i]=NPIECE;
  for (i=0; i<2*NPROPERTIES; i++)
    npiece[NPROPPOS+i]=(NPIECE*NPIECES)/2;
  
  RandomPivots();

  for (i=0; i<BOARDSIZE; i++) 
    XBlankPos(i);
  for (i=0; i<NPIECES; i++) 
    XInitPiece(i);

 for (i=0; i<NGAMETYPES; i++) {
    if (quarto_selected!=i)
      XtVaSetValues(gametype_button[i],XtNlabel,chosen_gametype_label[i],NULL);
    else
      XtVaSetValues(gametype_button[i],XtNlabel,gametype_label[i],NULL);
  }
  for (i=0; i<NLEVELS; i++) {
    if (maxdepth==i+1)
      XtVaSetValues(level_button[i],XtNlabel,chosen_level_label[i],NULL);
    else
      XtVaSetValues(level_button[i],XtNlabel,level_label[i],NULL);
  }
  for (i=0; i<NGEOMETRIES; i++) {
    if (geometry==i)
      XtVaSetValues(geometry_button[i],XtNlabel,chosen_geometry_label[i],NULL);
    else
      XtVaSetValues(geometry_button[i],XtNlabel,geometry_label[i],NULL);
  }

  if (realstart) {
    if (firstplayer==COMPUTER) {
      proposedpiece=pivotpiece[0];
      XPutOutPiece(proposedpiece);
      gamestatus=SELPOS;
      SetMessages(1,6);
    } else {   
      gamestatus=SELPIECE;
      if (firstplayer<TWOHUMANS)
        SetMessages(1,4);
      else
        SetMessages(2,4);
      proposedpiece=-1;
      XBlankOutPiece();
    }
  }

  ShowStatus();
}

static void PosProc(Widget widget,XtPointer posp,XtPointer call_data)
{
  int inpos=((PosPtr)posp)->posno;
  int* outpiece=((PosPtr)posp)->outpiece;

  if (gamestatus!=SELPOS) return;
  if (board[inpos]&1) return;

  if (firstplayer==NETWORK1) {
    if (widget==pos[inpos] && (move_num&1)==0) return;
    if (widget==pos2[inpos] && (move_num&1)==1) return;
  }
  if (firstplayer==NETWORK2) {
    if (widget==pos[inpos] && (move_num&1)==1) return;
    if (widget==pos2[inpos] && (move_num&1)==0) return;
  }

  addpiece(*outpiece,inpos);
  XAddPiece(*outpiece,inpos);
  seqpiece[move_num]=*outpiece;
  seqpos[move_num]=inpos;
  ++move_num;
  proposedpiece=-1;
  XBlankOutPiece();

  if (endgame()) {
    HilightProc();
    if(firstplayer<TWOHUMANS)
      SetMessages(0,(quarto_selected)? 9:10);
    else
    if(firstplayer<=NETWORK1)
      SetMessages(0,((move_num&1)==quarto_selected)? 12:11);
    else
      SetMessages(0,((move_num&1)==quarto_selected)? 11:12);

    gamestatus=ENDGAME;
    return;
  } else if (board[NPIECEPOS]==NPIECES) {
    SetMessages(0,13);
    gamestatus=ENDGAME;
    return;
  } else {
    if (firstplayer<TWOHUMANS)
      SetMessages(1,5);
    else {
    if (firstplayer<=NETWORK1)
      SetMessages((move_num&1)? 3:2, 5);
    else
      SetMessages((move_num&1)? 2:3, 5);
    }
    gamestatus=SELPIECE;
    return;  
  }
}

static void PieceProc(Widget widget,XtPointer piecep,XtPointer call_data)
{
  int inpiece=((PiecePtr)piecep)->pieceno;
  int* outpiece=((PiecePtr)piecep)->outpiece;
  int outpos=-1;

  if (gamestatus!=SELPIECE) return;

  if (firstplayer==NETWORK1) {
    if (widget==piece[inpiece] && (move_num&1)==1) return;
    if (widget==piece2[inpiece] && move_num==0) firstplayer=NETWORK2;
    else
    if (widget==piece2[inpiece] && (move_num&1)==0) return;
  }
  if (firstplayer==NETWORK2) {
    if (widget==piece[inpiece] && (move_num&1)==0) return;
    if (widget==piece2[inpiece] && (move_num&1)==1) return;
  }

  proposedpiece=inpiece;
  XPutOutPiece(inpiece);

  if (firstplayer<TWOHUMANS) {
    SetMessages(8,0);
    XFlushDpys();
    usleep(DELAY);
    XFlushDpys();

    if (selpospiece(inpiece,outpiece,&outpos)==-1) {
      SetMessages(0,9);
      gamestatus=ENDGAME;
      return;
    }

    usleep(DELAY);
    addpiece(inpiece,outpos);
    seqpiece[move_num]=inpiece;
    seqpos[move_num]=outpos;
    ++move_num;
    XAddPiece(inpiece,outpos);
    XBlankOutPiece();
    XFlushDpys();

  } else
      *outpiece = inpiece;

  if (quarto_selected && endgame()) {
    HilightProc();
    if(firstplayer<TWOHUMANS)
      SetMessages(0,10);
    else {
    if(firstplayer<=NETWORK1)
      SetMessages(0, (move_num&1)?11:12);
    else
      SetMessages(0, (move_num&1)?12:11);
    }
    proposedpiece=-1;
    gamestatus=ENDGAME;
    return;
  } else if (board[NPIECEPOS]==NPIECES) {
    SetMessages(0,13);
    proposedpiece=-1;
    gamestatus=ENDGAME;
    return;
  }
  if (*outpiece>=0 && *outpiece<=NPIECES) {
    if (firstplayer<TWOHUMANS) usleep(DELAY);
    proposedpiece=*outpiece;
    XPutOutPiece(*outpiece);
  }
  if (firstplayer<TWOHUMANS)
    SetMessages(1,7);
  else
  if (firstplayer<=NETWORK1)
    SetMessages((move_num&1)? 2:3, 7);
  else
    SetMessages((move_num&1)? 3:2, 7);
  gamestatus=SELPOS;
  return;
}

static void HilightProc()
{
  int i,j,hipos;
  long eg;

  eg=endgame();
  for (i=0; i<BOARDLENGTH; i++)   /* Check each row */
    if ((eg>>i)&1)
      for (j=0; j<BOARDLENGTH; j++) {
	hipos=i*BOARDLENGTH+j;
        XHilightPos(hipos);
      }
  for (i=0; i<BOARDLENGTH; i++)   /* Check each col */
    if ((eg>>(i+BOARDLENGTH))&1)
      for (j=0; j<BOARDLENGTH; j++) {
	hipos=j*BOARDLENGTH+i;
        XHilightPos(hipos);
      }
  for (i=0; i<2*BOARDLENGTH; i++)   /* Check each diag */
    if ((eg>>(i+2*BOARDLENGTH))&1)
      for (j=0; j<BOARDLENGTH; j++) {
	if (i&1)
	  hipos=j*BOARDLENGTH+((BOARDLENGTH+i/2-j)%BOARDLENGTH);
	else
	  hipos=j*BOARDLENGTH+((i/2+j)%BOARDLENGTH);
        XHilightPos(hipos);
      }
  return;
}

static void GametypeProc(Widget widget,XtPointer gametypep,XtPointer call_data)
{
  int newgametype=((GametypePtr)gametypep)->gametype;
  int i;
  quarto_selected = 1 - newgametype;
  for (i=0; i<NGAMETYPES; i++) {
    if (i==newgametype)
      XtVaSetValues(gametype_button[i],XtNlabel,chosen_gametype_label[i],NULL);
    else
      XtVaSetValues(gametype_button[i],XtNlabel,gametype_label[i],NULL);
  }
  InitGame();
  return;
}

static void LevelProc(Widget widget,XtPointer levelp,XtPointer call_data)
{
  int i, newlevel;

  newlevel=((LevelPtr)levelp)->level;
  maxdepth=newlevel+1;
  for (i=0; i<NLEVELS; i++) {
    if (maxdepth==i+1)
      XtVaSetValues(level_button[i],XtNlabel,chosen_level_label[i],NULL);
    else
      XtVaSetValues(level_button[i],XtNlabel,level_label[i],NULL);
  }
  ShowStatus();
  return;
}

static void GeometryProc(Widget widget,XtPointer geometryp,XtPointer call_data)
{
  int newgeometry=((GeometryPtr)geometryp)->geometry;
  int i;
  geometry=newgeometry;
  for (i=0; i<NGEOMETRIES; i++) {
    if (geometry==i)
      XtVaSetValues(geometry_button[i],XtNlabel,chosen_geometry_label[i],NULL);
    else
      XtVaSetValues(geometry_button[i],XtNlabel,geometry_label[i],NULL);
  }
  InitGame();
  return;
}

static void StartHumanProc(Widget widget,XtPointer client_data,XtPointer call_data)
{
  if (firstplayer>=NETWORK1) return;
  firstplayer=HUMAN;
  InitGame();
}

static void StartComputerProc(Widget widget,XtPointer client_data,XtPointer call_data)
{
  if (firstplayer>=NETWORK1) return;
  firstplayer=COMPUTER;
  InitGame();
}

static void StartTwohumansProc(Widget widget,XtPointer client_data,XtPointer call_data)
{
  if (firstplayer>=NETWORK1) return;
  firstplayer=TWOHUMANS;
  InitGame();
}

void StartNetworkProc(Widget w, XtPointer client_data, XtPointer call_data)
{
  Widget networkform, networktitle;

  if (networktop && dpy2) {
     firstplayer=NETWORK1;
     InitGame();
     SetMessages(14,4);
     return;
  }
  if (networktop) return;

  networktop = XtVaAppCreateShell(network_title, "networkshell",
       applicationShellWidgetClass, dpy, NULL);
  networkform = XtVaCreateManagedWidget("networkf",formWidgetClass,networktop,NULL);
  networktitle = XtVaCreateManagedWidget("networktitle",labelWidgetClass,networkform,
       XtNlabel, network_title,
       XtNvertDistance, 12,
       XtNborderWidth, 0,
       networkform,NULL);
  networkdismiss = XtVaCreateManagedWidget("networkdismiss",
       commandWidgetClass, networkform,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNhorizDistance, 457,
       XtNvertDistance, 10,           
       NULL);
  networklabel = XtVaCreateManagedWidget("networklabel",labelWidgetClass,networkform,
       XtNlabel, network_connect,
       XtNfromVert, networktitle,
       XtNwidth, 105, XtNheight, 20, 
       XtNvertDistance, 12,
       XtNborderWidth, 0,
       XtNjustify, 0,
       NULL);
  networksend = XtVaCreateManagedWidget("networkconnect",
       commandWidgetClass, networkform,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNfromVert, networklabel,
       XtNvertDistance, 20,           
       NULL);
  networkmsg = XtVaCreateManagedWidget("networkmsg",asciiTextWidgetClass,networkform,
       XtNstring, host,
       XtNfromVert, networktitle,
       XtNhorizDistance, 110,
       XtNbackground, white,
       XtNwidth, 417, XtNheight, 60, 
       XtNvertDistance, 12,
       XtNeditType, XawtextEdit,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded,
       NULL);
  networkmain =XtVaCreateManagedWidget("networkmain",asciiTextWidgetClass,networkform,
       XtNbackground, white,
       XtNwidth, 523,
       XtNheight, 207,
       XtNfromVert, networkmsg,
       XtNvertDistance, 10,
       XtNstring, "",
       XtNeditType, XawtextRead,
       XtNdisplayCaret, False,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded, NULL);
  XtAddCallback(networkdismiss,XtNcallback,
        CloseNetworkProc, (caddr_t) NULL);
  XtAddCallback(networksend,XtNcallback,
        MessageProc, (caddr_t) NULL); 
  XtRealizeWidget(networktop);
  XSetWMHints(dpy, XtWindow(networktop), hints);
  XtSetKeyboardFocus(networktop, networkmsg);
 
  firstplayer=TWOHUMANS;
  InitGame();
  SetMessages(0,0);
}

void CloseNetworkProc(Widget w, XtPointer client_data, XtPointer call_data)
{
  QuitProc2(w,client_data,call_data);
  XSetInputFocus(dpy,XtWindow(top),RevertToNone,CurrentTime);
  ShowStatus();
  SetMessages(1,4);
}

void StartNetworkProc2()
{
Widget networkform2, networktitle2;

  networktop2 = XtVaAppCreateShell(network_title2, "networkshell",
       applicationShellWidgetClass, dpy2, NULL);

  networkform2 = XtVaCreateManagedWidget("networkf",formWidgetClass,networktop2,NULL);
  networktitle2 = XtVaCreateManagedWidget("networktitle",labelWidgetClass,networkform2,
       XtNlabel, network_title,
       XtNvertDistance, 12,
       XtNborderWidth, 0,
       NULL);
  networkdismiss2 = XtVaCreateManagedWidget("networkdismiss",
       commandWidgetClass, networkform2,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNhorizDistance, 457,
       XtNvertDistance, 12,           
       NULL);
  networklabel2 = XtVaCreateManagedWidget("networklabel", labelWidgetClass,
       networkform2,
       XtNlabel, network_msg,
       XtNfromVert, networktitle2,
       XtNwidth, 105, XtNheight, 20, 
       XtNvertDistance, 12,
       XtNborderWidth, 0,
       XtNjustify, 0,
       NULL);
  networksend2 = XtVaCreateManagedWidget("networksend",
       commandWidgetClass, networkform2,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNfromVert, networklabel2,
       XtNvertDistance, 20,           
       NULL);
  networkmsg2 = XtVaCreateManagedWidget("networkmsg", asciiTextWidgetClass,
       networkform2,
       XtNstring, "",
       XtNfromVert, networktitle2,
       XtNhorizDistance, 110,
       XtNbackground, white2,
       XtNwidth, 417, XtNheight, 60, 
       XtNvertDistance, 12,
       XtNeditType, XawtextEdit,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded,
       NULL);
  networkmain2 =XtVaCreateManagedWidget("networkmain",asciiTextWidgetClass,networkform2,
       XtNbackground, white2,
       XtNwidth, 523,
       XtNheight, 207,
       XtNfromVert, networkmsg2,
       XtNvertDistance, 10,
       XtNstring, "",
       XtNeditType, XawtextRead,
       XtNdisplayCaret, False,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded, NULL);
  XtAddCallback(networksend2, XtNcallback, MessageProc, (caddr_t) NULL);
  XtAddCallback(networkdismiss2, XtNcallback, QuitProc2, (caddr_t) NULL);
  XtRealizeWidget(networktop2);
  XSetWMHints(dpy2, XtWindow(networktop2), hints);
  XtSetKeyboardFocus(networktop2, networkmsg2);
}

static void MessageProc(Widget widget,XtPointer client_data,
  XtPointer call_data)
{
char *oldtext, *oldtext2;
static char *newtext;
static char *text, *text2;
int i, j, length;

  if (dpy2) {
  if (widget==networksend) {
     i=1;
     XtVaGetValues(networkmsg, XtNstring, &newtext, NULL);
  }
  if (widget==networksend2) {
     i=2;
     XtVaGetValues(networkmsg2, XtNstring, &newtext, NULL);
  }
  length=strlen(newtext)-1;
  while((isspace(newtext[length]) || 
         iscntrl(newtext[length])) && length>=0) {
    newtext[length]='\0';
    --length;
  }
  if (length>=0) {
    XtVaGetValues(networkmain, XtNstring, &oldtext, NULL);
    XtVaGetValues(networkmain2, XtNstring, &oldtext2, NULL);
    length+=strlen(messages[18])+10;
    text=malloc((strlen(oldtext)+length)*sizeof(char));
    text2=malloc((strlen(oldtext2)+length)*sizeof(char));
    sprintf(text, "%s%s %d:\n%s\n\n", oldtext, messages[18], i, newtext);
    sprintf(text2, "%s%s %d:\n%s\n\n", oldtext2, messages[18], i, newtext);
    XtVaSetValues(networkmain, XtNstring, text, NULL);
    XawTextSetInsertionPoint(networkmain,strlen(text)-2);
    XtVaSetValues(networkmain2, XtNstring, text2, NULL);
    XawTextSetInsertionPoint(networkmain2,strlen(text2)-2);
  }
  if (widget==networksend)
     XtVaSetValues(networkmsg, XtNstring, "", NULL);
  if (widget==networksend2)
     XtVaSetValues(networkmsg2, XtNstring, "", NULL);
  return;
  }

  CreateWidgets2();
  if (dpy2==NULL) return;
  PrepareCallbacks2();
  StartNetworkProc2();
  firstplayer=NETWORK1;
  InitGame();
  SetMessages(14,4);
  text=malloc(400*sizeof(char));
  sprintf(text, messages[19], host);
  XtVaSetValues(networkmain, XtNstring, text, NULL);   
  sprintf(text, messages[20], getenv("USER"));
  XtVaSetValues(networkmain2, XtNstring, text, NULL);   
}

void HelpProc(Widget w, XtPointer client_data, XtPointer call_data)
{
  Widget helpform,helpmain,helptitle;
  /*
  XFontStruct *font;
  font = XLoadQueryFont(dpy, "-*-charter-bold-r-*-*-30-*-*-*-*-*-iso8859-*");
  */

  if (helptop)
    {
     XtMapWidget(helptop);
     XFlushDpys();
     return;
    }

  helptop = XtVaAppCreateShell(((PopupDefs*)client_data)->title, "helpshell",
       applicationShellWidgetClass, dpy, NULL);
  helpform = XtVaCreateManagedWidget("helpf",formWidgetClass,helptop,NULL);
  helptitle = XtVaCreateManagedWidget("helptitle",labelWidgetClass,helpform,
       XtNlabel, ((PopupDefs*)client_data)->title,
       XtNvertDistance, 12, /* XtNfont, font, */
       XtNborderWidth, 0,
       helpform,NULL);
  helpdismiss = XtVaCreateManagedWidget("helpdismiss",
       commandWidgetClass, helpform,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNhorizDistance, ((PopupDefs*)client_data)->distance,
       XtNvertDistance, 10,           
       NULL);
  helpmain =XtVaCreateManagedWidget("helpmain",asciiTextWidgetClass,helpform,
       XtNbackground, white,
       XtNwidth, ((PopupDefs*)client_data)->width,
       XtNheight, ((PopupDefs*)client_data)->height,
       XtNfromVert, helpdismiss,
       XtNvertDistance, 10,
       XtNstring, ((PopupDefs*)client_data)->string,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded, NULL);
  XtAddCallback(helpdismiss,XtNcallback, CloseHelpProc, (caddr_t) NULL);
  XtRealizeWidget(helptop);
  XSetWMHints(dpy, XtWindow(helptop), hints);
}

void CloseHelpProc(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtDestroyWidget(helptop);
  helptop=NULL;
  XSetInputFocus(dpy,XtWindow(top),RevertToNone,CurrentTime);
}

void HelpProc2(Widget w, XtPointer client_data, XtPointer call_data)
{
  Widget helpform2,helpmain2,helptitle2;

  helptop2 = XtVaAppCreateShell(((PopupDefs*)client_data)->title, 
          "helpshell",
       applicationShellWidgetClass, dpy2, NULL);
  helpform2 = XtVaCreateManagedWidget("helpf",formWidgetClass,helptop2,NULL);
  helptitle2 = XtVaCreateManagedWidget("helptitle",labelWidgetClass,helpform2,
       XtNlabel, ((PopupDefs*)client_data)->title,
       XtNvertDistance, 12, /* XtNfont, font, */
       XtNborderWidth, 0,
       helpform2,NULL);
  helpdismiss2 = XtVaCreateManagedWidget("helpdismiss",
       commandWidgetClass, helpform2,
       XtNwidth, 70,
       XtNheight, 20,
       XtNborderWidth, 1,
       XtNhorizDistance, ((PopupDefs*)client_data)->distance,
       XtNvertDistance, 10,           
       NULL);
  helpmain2 =XtVaCreateManagedWidget("helpmain",asciiTextWidgetClass,helpform2,
       XtNbackground, white2,
       XtNwidth, ((PopupDefs*)client_data)->width,
       XtNheight, ((PopupDefs*)client_data)->height,
       XtNfromVert, helpdismiss2,
       XtNvertDistance, 10,
       XtNstring, ((PopupDefs*)client_data)->string,
       XtNscrollHorizontal, XawtextScrollWhenNeeded,
       XtNscrollVertical,   XawtextScrollWhenNeeded, NULL);
  XtAddCallback(helpdismiss2,XtNcallback,
        CloseHelpProc2, (caddr_t) NULL);
  XtRealizeWidget(helptop2);
  XSetWMHints(dpy2, XtWindow(helptop2), hints);
}

void CloseHelpProc2(Widget w, XtPointer client_data, XtPointer call_data)
{
  XtDestroyWidget(helptop2);
  helptop2=NULL;
  XSetInputFocus(dpy2,XtWindow(top2),RevertToNone,CurrentTime);
}

static void QuitProc(Widget widget,XtPointer client_data,XtPointer call_data)
{
  exit(0);
}

static void NewProc2()
{
  firstplayer=NETWORK1;
  InitGame();
  SetMessages(14,4);
}

static void QuitProc2()
{
  if (top2)
    XtDestroyWidget(top2);
  if (helptop2)
    XtDestroyWidget(helptop2);
  if (networktop2)
    XtDestroyWidget(networktop2);
  if (networktop) 
    XtDestroyWidget(networktop);
  dpy2=NULL;
  top2=NULL;
  helptop2=NULL;
  networktop2=NULL;
  networktop=NULL;
  firstplayer=HUMAN;
  InitGame();
  SetMessages(15,4);
  ShowStatus();
}

static void SaveProc()
{
FILE *fd;
int  i;

  fd=fopen(savefile,"w");
  if(fd) {
    fprintf(fd,"Status: %s\n", statusline);
    fprintf(fd,"Proposed: %d\n", proposedpiece);
    for (i=0; i<move_num; i++)
      fprintf(fd,"Step %2d: %2d -> %c%c\n", i+1, seqpiece[i], 
                 'A'+seqpos[i]/4, '1'+seqpos[i]%4);
    fclose(fd);
  } else {
    if (helptop) CloseHelpProc(NULL, NULL, NULL);
    sprintf(message_body, alert[1], savefile, alert[3]);
    HelpProc(NULL, (XtPointer)&AlertDefs, NULL);
  }
}

static void ProcessRecord()
{
int i, j, k;

  for(i=0; i<move_num; i++) {
    j=seqpiece[i];
    k=seqpos[i];
    addpiece(j, k);
    XAddPiece(j,k);
    XFadePiece(j);
  }
  if (proposedpiece<0) {
    gamestatus=SELPIECE;
    XBlankOutPiece();
  } else {
    gamestatus=SELPOS;
    XPutOutPiece(proposedpiece);
  }
  if (proposedpiece<0)
     i=move_num&1;
  else
     i=1-move_num&1;
  if(firstplayer==HUMAN) SetMessages((i)? 0:1,(proposedpiece<0)? 5:7);
  else
  if(firstplayer==COMPUTER) SetMessages((i)? 1:0,(proposedpiece<0)? 5:7);
  else
  if(firstplayer<=NETWORK1) SetMessages((i)? 3:2,(proposedpiece<0)? 5:7);
  else
  if(firstplayer==NETWORK2) SetMessages((i)? 2:3,(proposedpiece<0)? 5:7);
  if (endgame()) {
    HilightProc();
    gamestatus=ENDGAME;
    if(!quarto_selected) i=!i;
    if(firstplayer==HUMAN) SetMessages(0,(i)? 10:9);
    else
    if(firstplayer==COMPUTER) SetMessages(0,(i)? 9:10);
    else
    if(firstplayer<=NETWORK1) SetMessages(0,(i)? 12:11);
    else
    if(firstplayer==NETWORK2) SetMessages(0,(i)? 11:12);
  } else
    if (board[NPIECEPOS]==NPIECES) SetMessages(0,13);
}

static void LoadProc()
{
FILE *fd;
char *ptr, c, d, line[512];
char players[6], game_type[10], ch_level, geom_type[10];
int  i, j;

  if(networktop && dpy2==NULL) return;

  fd=fopen(savefile,"r");
  if(fd) {
    ptr=fgets(line,80,fd);
    if(ptr) {
      for (i=0; i<strlen(line); i++) if (line[i]=='/') line[i]=' ';
      sscanf(line,"Status: %s%s level%c %s", 
              players, game_type, &ch_level, geom_type);
    }
    if (firstplayer<NETWORK1 && 
        (!strcmp(players,"H:N") || !strcmp(players,"N:H"))) {
          SetMessages(16,-1);
          fclose(fd);
          return;
    }
    if (firstplayer>=NETWORK1 && 
        (strcmp(players,"H:N") && strcmp(players,"N:H"))) {
          SetMessages(17,-1);
          fclose(fd);
          return;
    }
    if (!strcmp(players,"H:C")) firstplayer=HUMAN;
    if (!strcmp(players,"C:H")) firstplayer=COMPUTER;
    if (!strcmp(players,"H:H")) firstplayer=TWOHUMANS;
    if (!strcmp(players,"H:N")) firstplayer=NETWORK1;
    if (!strcmp(players,"N:H")) firstplayer=NETWORK2;
    quarto_selected=!strcmp(game_type,"quarto");
    maxdepth=ch_level-'0';
    if (!strcmp(geom_type,"normal")) geometry=NORMAL;
    if (!strcmp(geom_type,"nodiags")) geometry=NODIAGS;
    if (!strcmp(geom_type,"torus")) geometry=TORUS;
    realstart=0;
    InitGame();
    realstart=1;
    ptr=fgets(line,80,fd);
    if(ptr) {
      sscanf(line,"Proposed: %d", &proposedpiece);
    }
    while((ptr=fgets(line,80,fd))!=NULL) {
      if(sscanf(line,"Step %d: %d -> %c%c\n", &i, &j, &c, &d)==4) {
        move_num=i;
        --i;
        seqpiece[i]=j;
        seqpos[i]=4*(c-'A')+(d-'1');
      }
    }
    fclose(fd);
    ProcessRecord();
  } else {
    if (helptop) CloseHelpProc(NULL, NULL, NULL);
    sprintf(message_body, alert[2], savefile, alert[3]);
    HelpProc(NULL, (XtPointer)&AlertDefs, NULL);
  }
}

static void UndoProc(Widget widget,XtPointer posp,XtPointer call_data)
{
int i,j,k;

  if (firstplayer==HUMAN) {  
    if (move_num>0) {
      i=move_num-1;
      if (move_num&1) 
        j=-1;
      else
        j=seqpiece[i];
    } else 
      return;
  }
  if (firstplayer==COMPUTER) {  
    if (move_num>0) {
      i=move_num-1;
      if (move_num&1==0) 
        j=-1;
      else
        j=seqpiece[i];
    } else 
      return;
  }
  k=(widget==undo_button);
  if (firstplayer>=TWOHUMANS) {  
    if (proposedpiece>=0) {
      if (firstplayer==NETWORK1 && (move_num&1)==k) return;
      if (firstplayer==NETWORK2 && (move_num&1)!=k) return;
      i=move_num;
      j=-1;
    } else {
      if (move_num==0) 
        return;
      else {
        if (firstplayer==NETWORK1 && (move_num&1)==k) return;
        if (firstplayer==NETWORK2 && (move_num&1)!=k) return;
        i=move_num-1;
        j=seqpiece[i];
      }
    }
  }
  realstart=0;
  InitGame();
  realstart=1;
  move_num=i;
  proposedpiece=j;
  ProcessRecord();
}

static void HintProc()
{
  int inpiece, outpiece, outpos, i;

  outpos = -1;
  XFlushDpys();

  RandomPivots();

  if (gamestatus==SELPIECE) {
     inpiece = -1;
     if (move_num)
        selpospiece(inpiece, &outpiece, &outpos);
     else
        outpiece=pivotpiece[0];
     if(outpiece<0) return;
     for (i=0; i<8; i++) {
         XBlankPiece(outpiece);
         XFlushDpys();
         usleep(DELAY/10);
         XInitPiece(outpiece);
         XFlushDpys();
         usleep(DELAY/10);
     }
  }

  if (gamestatus==SELPOS) {
     if (proposedpiece==-1) 
       return;
     else
       inpiece=proposedpiece;
     selpospiece(inpiece, &outpiece, &outpos);
     if(inpiece<0 || outpos<0) return;
     for (i=0; i<8; i++) {
         XAddPiece(inpiece,outpos);
         XFlushDpys();
         usleep(DELAY/10);
         XBlankPos(outpos);
         XFlushDpys();
         usleep(DELAY/10);
     }
  }
}

